/*
 * Open Source Physics software is free software as described near the bottom of this code file.
 *
 * For additional information and documentation on Open Source Physics please see:
 * <http://www.opensourcephysics.org/>
 */

package org.opensourcephysics.tools;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.text.DecimalFormat;
import java.text.NumberFormat;

import javax.swing.*;
import javax.swing.undo.*;

import org.opensourcephysics.controls.*;
import org.opensourcephysics.numerics.*;

/**
 * A dialog for editing user functions. 
 * 
 * @author Douglas Brown
 */
public class UserFunctionEditor extends JDialog 
		implements PropertyChangeListener {
	
	final static Color LIGHT_RED = new Color(255, 127, 127);
	final static Color LIGHT_MAGENTA = new Color(255, 153, 204);

	protected ParameterPanel paramPanel;
	protected UserFunction function;
	protected DatasetCurveFitter fitter;
	protected XMLTable table;
	protected boolean deleted;
	protected UndoableEditSupport undoSupport;
	protected UndoManager undoManager;
	protected JButton undoButton;
	protected JButton redoButton;
  protected static NumberFormat format = NumberFormat.getNumberInstance();
  
  static {
    ((DecimalFormat)format).applyPattern("0.000E0"); //$NON-NLS-1$
  }

	/**
	 * Constructor.
	 * 
	 * @param f the user function to edit
	 */
	public UserFunctionEditor(UserFunction f) {
		this (f, null);
	}
	
	/**
	 * Constructor.
	 * 
	 * @param func the user function to edit
	 */
	public UserFunctionEditor(UserFunction func, DatasetCurveFitter fitter) {
		super(fitter != null && fitter.getTopLevelAncestor() instanceof JFrame? 
						(JFrame)fitter.getTopLevelAncestor(): (JFrame)null, true); // modal
		function = func;
		this.fitter = fitter;
		// set up the undo system
		undoManager = new UndoManager();
		undoSupport = new UndoableEditSupport();
		undoSupport.addUndoableEditListener(undoManager);
		// create the GUI
		setTitle(ToolsRes.getString("UserFunctionEditor.Title")  //$NON-NLS-1$ 
						+ " \"" + func.getName() + "\""); //$NON-NLS-1$ //$NON-NLS-2$
		paramPanel = new ParameterPanel();
		paramPanel.addPropertyChangeListener("parameter", this); //$NON-NLS-1$
		paramPanel.getTable().addMouseListener(new MouseAdapter() {
      public void mousePressed(MouseEvent e) {
      	AbstractCellEditor editor = (AbstractCellEditor)table.getCellEditor();
      	if (editor != null) editor.stopCellEditing();
      }
    });
		for (int i = 0; i < func.getParameterCount(); i++) {
			paramPanel.addParameter(new Parameter(
							func.getParameterName(i), String.valueOf(func.getParameterValue(i))));
		}
		// put vertical box in CENTER with parameter table
		Box box = Box.createVerticalBox();
		getContentPane().add(box, BorderLayout.CENTER);
		box.add(paramPanel);
		// add titled border to paramPanel
		paramPanel.setBorder(BorderFactory.createTitledBorder(
						ToolsRes.getString("UserFunctionEditor.Parameter.Border.Title"))); //$NON-NLS-1$
    // make new panel with XMLTable in scroller
    XMLControlElement control = new XMLControlElement();
    control.setValue(func.getName(), func.getExpression());
    table = new OSPControlTable(control);
    table.setEditable(true);
    table.addPropertyChangeListener("tableData", this); //$NON-NLS-1$
    table.addMouseListener(new MouseAdapter() {
      public void mousePressed(MouseEvent e) {
      	AbstractCellEditor editor = 
      		(AbstractCellEditor)paramPanel.getTable().getCellEditor();
      	if (editor != null) editor.stopCellEditing();
      }
    });
		JScrollPane scroller = new JScrollPane(table);
		scroller.createHorizontalScrollBar();
		JPanel xmlPanel = new JPanel(new BorderLayout()) {
			// override getPreferredSize to report table size
			public Dimension getPreferredSize() {
				Dimension dim = table.getPreferredSize();
				return dim;
			}	
		};
		xmlPanel.add(scroller, BorderLayout.CENTER);
		// add titled border to xmlPanel
		xmlPanel.setBorder(BorderFactory.createTitledBorder(
				ToolsRes.getString("UserFunctionEditor.Function.Border.Title") //$NON-NLS-1$
				+ " ("+func.getIndependentVariable()+")")); //$NON-NLS-1$ //$NON-NLS-2$
		box.add(xmlPanel);
		// put buttonBar in SOUTH position
		JPanel buttonBar = new JPanel(new FlowLayout());
		getContentPane().add(buttonBar, BorderLayout.SOUTH);
		// undo button
		undoButton = new JButton(ToolsRes.getString("UserFunctionEditor.Button.Undo")); //$NON-NLS-1$
		undoButton.setEnabled(false);
		buttonBar.add(undoButton);
		undoButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				undoManager.undo();
			}
		});
		// redo button
		redoButton = new JButton(ToolsRes.getString("UserFunctionEditor.Button.Redo")); //$NON-NLS-1$
		redoButton.setEnabled(false);
		buttonBar.add(redoButton);
		redoButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				undoManager.redo();
			}
		});
		// delete button
		JButton deleteButton = new JButton(
						ToolsRes.getString("UserFunctionEditor.Button.Delete") //$NON-NLS-1$
						+" "+func.getName()); //$NON-NLS-1$
		buttonBar.add(deleteButton);
		deleteButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				deleted = true;
				setVisible(false);
			}
		});
		// close button
		JButton closeButton = new JButton(ToolsRes.getString("UserFunctionEditor.Button.Close")); //$NON-NLS-1$
		buttonBar.add(closeButton);
		closeButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				// set user function parameters
				function.setParameters(paramPanel.getNames(), paramPanel.getValues());
				// set user function equation
				XMLControl control = table.getControl();
				String eqn = control.getString(function.getName());
				function.setExpression(eqn, function.getIndependentVariable());
				setVisible(false);
			}
		});
		pack();
		setSize(getWidth(), 280);
    Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
    int x = (dim.width - getBounds().width) / 2;
    int y = (dim.height - getBounds().height) / 2;
    setLocation(x, y);
	}
	
	/**
	 * Returns true if the user has deleted the user function.
	 * 
	 * @return true if deleted
	 */
	public boolean isDeleted() {
		return deleted;
	}
	
	/**
	 * Refreshes the display based on the current function values.
	 */
	public void refresh() {
		Parameter[] params = paramPanel.getParameters();
		for (int i = 0; i < function.getParameterCount(); i++) {
			// post undo edits if parameters have changed
			if (params[i].getValue() != function.getParameterValue(i)) {
				paramPanel.replaceParameter(new Parameter(params[i].getName(), 
								format.format(function.getParameterValue(i))));
				Parameter[] newParams = paramPanel.getParameters();
				ParameterPanel.ParamEdit edit = paramPanel.new ParamEdit(
								newParams, params, i, i, 1, 1);
				undoSupport.postEdit(edit);
				params = newParams;
			}
		}
	}
	
	/**
	 * Responds to property change events.
	 * Receives "parameter" from ParameterPanel and "tableData" from XMLTable.
	 * 
	 * @param e the property change event
	 */
	public void propertyChange(PropertyChangeEvent e) {
		// get current equation
		String eqn = table.getControl().getString(function.getName());
		if ("".equals(eqn)) { //$NON-NLS-1$
			eqn = "0"; //$NON-NLS-1$
			table.getControl().setValue(function.getName(), eqn);
		}
		// post undo edit and refresh buttons
		if (e != null && e.getNewValue() instanceof UndoableEdit) {
			undoSupport.postEdit((UndoableEdit)e.getNewValue());
		}
		else if (e != null && e.getPropertyName().equals("tableData")) { //$NON-NLS-1$
			if (!function.getExpression().equals(eqn)) {
				UndoableEdit edit = new FunctionEdit(eqn, function.getExpression());
				undoSupport.postEdit(edit);
			}
		}
		undoButton.setEnabled(undoManager.canUndo());
		redoButton.setEnabled(undoManager.canRedo());
		// try to parse equation
		eqn = eqn.replaceAll(function.getIndependentVariable(), "dummy"); //$NON-NLS-1$
		String[] params = paramPanel.getNames();
		String[] names = new String[params.length+1];
		names[0] = "dummy"; //$NON-NLS-1$
		System.arraycopy(params, 0, names, 1, params.length);
		boolean error = eqn.indexOf("=") > -1; //$NON-NLS-1$
		try {
			new ParsedMultiVarFunction(eqn, names);
		}
		catch (ParserException ex) {error = true;}
		if (error) {
			table.setBackgroundColor(function.getName(), LIGHT_RED);
			table.setSelectedColor(function.getName(), LIGHT_MAGENTA);
		}
		else {
			table.setBackgroundColor(function.getName(), null);
			table.setSelectedColor(function.getName(), null);
			// refresh fitter
			if (fitter != null) {
				// set user function parameters
				function.setParameters(paramPanel.getNames(), paramPanel.getValues());
				// set user function equation
				function.setExpression(eqn, function.getIndependentVariable());
				fitter.selectFit(function.getName());
				fitter.refreshGUI();
			}
		}
		table.repaint();
	}
	
	/**
	 * A class to undo/redo function expression changes.
	 */
	class FunctionEdit extends AbstractUndoableEdit {

		String prev;
		String curr;

		// constructor
		public FunctionEdit(String current, String previous) {
			prev = previous;
			curr = current;
		}

		// undoes the change
		public void undo() throws CannotUndoException {
			super.undo();
			table.getControl().setValue(function.getName(), prev);
			propertyChange(null);
		}

		// redoes the change
		public void redo() throws CannotUndoException {
			super.redo();
			table.getControl().setValue(function.getName(), curr);
			propertyChange(null);
		}

		// returns the presentation name
		public String getPresentationName() {
			return "Function"; //$NON-NLS-1$
		}
	}
}

/*
 * Open Source Physics software is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License (GPL) as
 * published by the Free Software Foundation; either version 2 of the License,
 * or(at your option) any later version.

 * Code that uses any portion of the code in the org.opensourcephysics package
 * or any subpackage (subdirectory) of this package must must also be be released
 * under the GNU GPL license.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA
 * or view the license online at http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (c) 2007  The Open Source Physics project
 *                     http://www.opensourcephysics.org
 */
